/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.lang.ri.executor;

import java.util.ArrayList;
import java.util.List;
import javax.swing.event.EventListenerList;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Properties;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.lang.ExceptionEvent;
import org.jdtaus.core.lang.ExceptionListener;
import org.jdtaus.core.lang.spi.Executor;

/**
 * jDTAUS Core SPI JDK 1.4 {@code Executor} reference implementation.
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: Jdk14Executor.java 4556 2007-12-26 22:19:43Z schulte2005 $
 *
 * @see org.jdtaus.core.container.Container
 * @see #initialize()
 */
public final class Jdk14Executor
    implements Executor, ContainerInitializer
{
    //--Implementation----------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausImplementation
    // This section is managed by jdtaus-container-mojo.

    /** Meta-data describing the implementation. */
    private static final Implementation META =
        ModelFactory.getModel().getModules().
        getImplementation(Jdk14Executor.class.getName());
// </editor-fold>//GEN-END:jdtausImplementation

    //----------------------------------------------------------Implementation--
    //--Constructors------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausConstructors
    // This section is managed by jdtaus-container-mojo.

    /**
     * <code>Jdk14Executor</code> implementation constructor.
     *
     * @param meta Implementation meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    private Jdk14Executor(final Implementation meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }
    /**
     * <code>Jdk14Executor</code> dependency constructor.
     *
     * @param meta dependency meta-data.
     *
     * @throws NullPointerException if <code>meta</code> is <code>null</code>.
     */
    private Jdk14Executor(final Dependency meta)
    {
        super();
        if(meta == null)
        {
            throw new NullPointerException("meta");
        }
        this.initializeProperties(meta.getProperties());
    }

    /**
     * Initializes the properties of the instance.
     *
     * @param meta the property values to initialize the instance with.
     *
     * @throws NullPointerException if {@code meta} is {@code null}.
     */
    private void initializeProperties(final Properties meta)
    {
        Property p;

        if(meta == null)
        {
            throw new NullPointerException("meta");
        }

        p = meta.getProperty("threadGroupName");
        this._threadGroupName = (java.lang.String) p.getValue();

    }
// </editor-fold>//GEN-END:jdtausConstructors

    //------------------------------------------------------------Constructors--
    //--ContainerInitializer----------------------------------------------------

    /**
     * {@code ThreadGroup} of threads executing the given {@code Runnable}s
     * asynchronously.
     */
    private ThreadGroup threadGroup;

    /**
     * Initializes the instance.
     *
     * @throws PropertyException for invalid property values.
     *
     * @see #assertValidProperties()
     */
    public void initialize()
    {
        this.assertValidProperties();
        this.threadGroup = new Jdk14ThreadGroup(this.getThreadGroupName());
    }

    //----------------------------------------------------ContainerInitializer--
    //--Dependencies------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausDependencies
    // This section is managed by jdtaus-container-mojo.

// </editor-fold>//GEN-END:jdtausDependencies

    //------------------------------------------------------------Dependencies--
    //--Properties--------------------------------------------------------------

// <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:jdtausProperties
    // This section is managed by jdtaus-container-mojo.

    /**
     * Property {@code threadGroupName}.
     * @serial
     */
    private java.lang.String _threadGroupName;

    /**
     * Gets the value of property <code>threadGroupName</code>.
     *
     * @return the value of property <code>threadGroupName</code>.
     */
    public java.lang.String getThreadGroupName()
    {
        return this._threadGroupName;
    }

// </editor-fold>//GEN-END:jdtausProperties

    //--------------------------------------------------------------Properties--
    //--ExceptionEventSource----------------------------------------------------

    /** Holds {@code ExceptionListener}s. */
    private final EventListenerList listeners = new EventListenerList();

    public void addExceptionListener(final ExceptionListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        synchronized(this.listeners)
        {
            this.listeners.add(ExceptionListener.class, listener);
        }
    }

    public void removeExceptionListener(final ExceptionListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        synchronized(this.listeners)
        {
            this.listeners.remove(ExceptionListener.class, listener);
        }
    }

    public ExceptionListener[] getExceptionListeners()
    {
        synchronized(this.listeners)
        {
            return (ExceptionListener[]) this.listeners.getListeners(
                ExceptionListener.class);

        }
    }

    //----------------------------------------------------ExceptionEventSource--
    //--ExceptionHandler--------------------------------------------------------

    /** List of exception handled by the instance. */
    private final List handledExceptions = new ArrayList(255);

    public ExceptionEvent[] getExceptionEvents()
    {
        synchronized(this.handledExceptions)
        {
            return (ExceptionEvent[]) this.handledExceptions.toArray(
                new ExceptionEvent[this.handledExceptions.size()]);

        }
    }

    public void handle(final Throwable t)
    {
        if(t == null)
        {
            throw new NullPointerException("t");
        }

        final ExceptionEvent evt =
            new ExceptionEvent(this, Thread.currentThread(), t);

        this.fireOnException(evt);
    }

    //--------------------------------------------------------ExceptionHandler--
    //--Executor----------------------------------------------------------------

    public void executeAsynchronously(final Runnable runnable)
    {
        if(runnable == null)
        {
            throw new NullPointerException("runnable");
        }

        new Thread(this.threadGroup, runnable).start();
    }

    //----------------------------------------------------------------Executor--
    //--Jdk14Executor-----------------------------------------------------------

    /**
     * Group of threads executing the given {@code Runnable}s asynchronously.
     */
    private final class Jdk14ThreadGroup extends ThreadGroup
    {
        /**
         * Creates a new {@code Jdk14ThreadGroup} instance taking its name.
         *
         * @param name the name of the new thread group.
         */
        Jdk14ThreadGroup(final String name)
        {
            super(name);
        }

        /**
         * {@inheritdDoc}
         * <p>This method notifies any registered {@code ExceptionListener}s
         * about the given exception if it is not an instance of
         * {@code ThreadDeath}. For instances of {@code ThreadDeath} the super
         * classe's {@code uncaughtException(Thread, Throwable)} method is
         * called.</p>
         *
         * @param t the thread that is about to exit.
         * @param e the uncaught exception.
         */
        public void uncaughtException(final Thread t, final Throwable e)
        {
            if(e instanceof ThreadDeath)
            {
                super.uncaughtException(t, e);
            }
            else
            {
                fireOnException(new ExceptionEvent(Jdk14Executor.this, t, e));
            }
        }
    };

    /** Creates a new {@code Jdk14Executor} instance. */
    public Jdk14Executor()
    {
        this(META);
        this.initialize();
    }

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for illegal property values.
     */
    private void assertValidProperties()
    {
        if(this.getThreadGroupName() == null ||
            this.getThreadGroupName().length() <= 0)
        {
            throw new PropertyException("threadGroupName",
                this.getThreadGroupName());

        }
    }

    /**
     * Notifies all registered {@code ExceptionListener}s about a given
     * {@code ExceptionEvent}.
     *
     * @param e the event to provide to the listeners.
     *
     * @throws NullPointerException if {@code e} is {@code null}.
     */
    private void fireOnException(final ExceptionEvent e)
    {
        if(e == null)
        {
            throw new NullPointerException("e");
        }

        synchronized(this.handledExceptions)
        {
            this.handledExceptions.add(e);
        }

        synchronized(this.listeners)
        {
            final Object[] listeners = this.listeners.getListenerList();
            for (int i = listeners.length - 2; i >= 0; i -= 2)
            {
                if (listeners[i] == ExceptionListener.class)
                {
                    ((ExceptionListener) listeners[i + 1]).onException(e);
                }
            }
        }
    }

    //-----------------------------------------------------------Jdk14Executor--
}
